// -*- C++ -*-
/*!
 * @file  MotionGenerator.cpp * @brief Motion generator component for reference hardware arm modle * $Date$ 
 *
 * $Id$ 
 */
#include "MotionGenerator.h"

// Module specification
// <rtc-template block="module_spec">
static const char* motiongenerator_spec[] =
  {
    "implementation_id", "MotionGenerator",
    "type_name",         "MotionGenerator",
    "description",       "Motion generator component for reference hardware arm modle",
    "version",           "1.0",
    "vendor",            "Junichiro Ooga, TOSHIBA",
    "category",          "RobotController",
    "activity_type",     "PERIODIC",
    "kind",              "DataFowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    ""
  };
// </rtc-template>

MotionGenerator::MotionGenerator(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_TargetVelocityOut("TargetVelocity", m_TargetVelocity),
    m_CurrentAngleIn("CurrentAngle", m_CurrentAngle),
    m_TargetAngleIn("TargetAngle", m_TargetAngle),
    m_TargetPositionIn("TargetPosition", m_TargetPosition)

    // </rtc-template>
{
}

MotionGenerator::~MotionGenerator()
{
}


RTC::ReturnCode_t MotionGenerator::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("CurrentAngle", m_CurrentAngleIn);
  addInPort("TargetAngle", m_TargetAngleIn);
  addInPort("TargetPosition", m_TargetPositionIn);

  // Set OutPort buffer
  addOutPort("TargetVelocity", m_TargetVelocityOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable

  // </rtc-template>

  L[0] = 0.125+0.125; // #0 -> #1
  L[1] = 0.0;         // #1 -> #2
  L[2] = 0.250;       // #2 -> #3
  L[3] = 0.110+0.09;  // #3 -> #4
  L[4] = 0.0;         // #4 -> #5
  L[5] = 0.0;         // #5 -> #6

  MAX_L = L[0] + L[1] + L[2] + L[3] + L[4] + L[5];
  
  
  Te[0][0] = 1.0;  Te[0][1] = 0.0;   Te[0][2] = 0.0;   Te[0][3] = 0.0;
  Te[1][0] = 0.0;  Te[1][1] = 1.0;   Te[1][2] = 0.0;   Te[1][3] = 0.0;
  Te[2][0] = 0.0;  Te[2][1] = 0.0;   Te[2][2] = 1.0;   Te[2][3] = 0.055+0.0805;

  CalcInvHomoMtrx(Te, inv_Te);

  m_TargetVelocity.data.length(JOINTNUM+1);

  PosLog.open("log_pos.dat");
  VelLog.open("log_vel.dat");
  AccLog.open("log_acc.dat");

  JPosLog.open("log_Jpos.dat");
  JVelLog.open("log_Jvel.dat");
  
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t MotionGenerator::onFinalize()
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t MotionGenerator::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t MotionGenerator::onShutdown(RTC::UniqueId ec_id)
{
  PosLog.close();
  VelLog.close();
  AccLog.close();
  JPosLog.close();
  JVelLog.close();

  std::cout << "Finish onShutdown" << std::endl;
  return RTC::RTC_OK;
}


RTC::ReturnCode_t MotionGenerator::onActivated(RTC::UniqueId ec_id)
{
  timeval tmpTime;
  gettimeofday(&tmpTime, NULL);
  OffsetTime = double(tmpTime.tv_sec + tmpTime.tv_usec * 1.0e-6);  
  NowTime = 0.0;
  
  iFlgCalcTrajParam = 0;
  
  // calculation of status valiables
  if(m_CurrentAngleIn.isNew()){

    m_CurrentAngleIn.read();
    // initialization of variables
    for(int i=0; i<JOINTNUM; i++){
    
      CurJPos[i] = m_CurrentAngle.data[i];
      OldJPos[i] = m_CurrentAngle.data[i];
      CurJVel[i] = 0.0;
      OldJVel[i] = 0.0;
      CurJAcc[i] = 0.0;
      IntegralJPos[i] = 0.0;
      
      RefJPos[i] = m_CurrentAngle.data[i];
      // RefVel[i] = 0.0;
      // RefAcc[i] = 0.0;

      m_TargetVelocity.data[i]= 0.0;
    }
  }else{
    // initialization of variables
    for(int i=0; i<JOINTNUM; i++){
    
      CurJPos[i] = 0.0;
      OldJPos[i] = 0.0;
      CurJVel[i] = 0.0;
      OldJVel[i] = 0.0;
      CurJAcc[i] = 0.0;
      IntegralJPos[i] = 0.0;

      RefJPos[i] = 0.0;

      m_TargetVelocity.data[i]= 0.0;
    }
  }

  if(m_TargetAngleIn.isNew()){
    m_TargetAngleIn.read();
    m_TargetAngle.data.length(JOINTNUM);
    for(int i=0; i<JOINTNUM; i++){
      LastJPos[i] = m_TargetAngle.data[i];
    }
    iFlgCalcTrajParam = 1;
  }

  ForwardKinematics(CurJPos, T, R);    // forward kinematics
  double P[6];
  CalcPositionVector(R[JOINTNUM], P);  

  for(int i=0; i<JOINTNUM; i++){
    
    a[i][4] = 0.0;
    a[i][3] = 0.0;
    a[i][2] = 0.0;
    a[i][1] = 0.0;
    a[i][0] = 0.0;
  }

  if(m_TargetPositionIn.isNew()){
    m_TargetPosition.data.length(JOINTNUM);
    m_TargetPositionIn.read();
    for(int i=0; i<6; i++){
      RefPos[i] = m_TargetPosition.data[i];
    }
  }else{
    for(int i=0; i<6; i++){
      RefPos[i] = P[i];
    }    
  }
  for(int i=0; i<JOINTNUM; i++){
    m_TargetVelocity.data[i] = (RefJPos[i] - CurJPos[i])*10.0;
  }
  m_TargetVelocity.data[JOINTNUM] = 0.0;
  m_TargetVelocityOut.write();
  
  t = 0.0;
  std::cout << "Finish onActivated" << std::endl;  
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t MotionGenerator::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t MotionGenerator::onExecute(RTC::UniqueId ec_id)
{
  int i,j,k;
  double Tw[3][4], Tw0[3][4], Jw[6][JOINTNUM], J11[3][JOINTNUM-3], J22[3][3];
  double drw[3], dww[3], dtheta[JOINTNUM];
  double Ti[JOINTNUM][3][4], Ri[JOINTNUM+1][3][4];

  double T0[3][4], Tn[3][4], tmpT[3][4], dTw[3][4];
    
  OldTime = NowTime;
  timeval tmpTime;
  gettimeofday(&tmpTime, NULL);

  NowTime = double(tmpTime.tv_sec + tmpTime.tv_usec * 1.0e-6)-OffsetTime;

  double dT = NowTime - OldTime;

  if(m_CurrentAngleIn.isNew()){
    m_CurrentAngle.data.length(JOINTNUM);
    m_CurrentAngleIn.read();
    // update valiables
    for(i=0; i<JOINTNUM; i++){
      OldJPos[i] = CurJPos[i];
      CurJPos[i] = m_CurrentAngle.data[i];
      //      IntegralPos[i] = 0.0;
      OldJVel[i] = CurJVel[i];
      CurJVel[i] = (CurJPos[i]-OldJPos[i])/dT;
      CurJAcc[i] = (CurJVel[i]-OldJVel[i])/dT;
    }    
  }

  TIME = 5.0; // <- configuration?

  if(m_TargetAngleIn.isNew()){
    m_TargetAngle.data.length(JOINTNUM);
    m_TargetAngleIn.read();
    for(int i=0; i<JOINTNUM; i++){
      LastJPos[i] = m_TargetAngle.data[i];
    }
    std::cout << "=================================================" << std::endl;
    iFlgCalcTrajParam = 1;    
  }else
  if(m_TargetPositionIn.isNew()){
    m_TargetPosition.data.length(6);
    m_TargetPositionIn.read();

    ForwardKinematics(CurJPos, T, R);  //  forward kinematics
    CalcPositionVector(R[JOINTNUM], CurPos);

    double zero[6];
    for(i=0; i<6; i++){
      LastPos[i] = m_TargetPosition.data[i];
      zero[i] = 0.0;
    }

    
    double Ln = CalcDistance(LastPos, zero);

    std::cout << "Pn=[" << LastPos[0] << ", " << LastPos[1] << ", " << LastPos[2] << ", "
	      << LastPos[3] << ", " << LastPos[4] << ", " << LastPos[5] << "], Ln=" << Ln << std::endl;

    if(Ln > MAX_L || Ln < L[0]){
      std::cout << "[out of range] Cannot reach this point!" << std::endl;
      return RTC::RTC_OK;
    }else{

      iFlgCalcTrajParam = 1;
      
      std::cout << "P0=[" << CurPos[0] << ", " << CurPos[1] << ", " << CurPos[2] << ", "
	      << CurPos[3] << ", " << CurPos[4] << ", " << CurPos[5] << "]" << std::endl;
      double distance = CalcDistance(LastPos, CurPos);
      std::cout << "distance = " << distance << std::endl; // for debug
      double dL = 0.01; // 10mm    
      int secNum = (int)(distance / dL ) + 1;
      
      CalcHomoMtrx(LastPos, Tn);
      CopyHomoMtrx(R[JOINTNUM], T0);
      CalcInvHomoMtrx(T0, tmpT);
      double Rot[3][4];              // T0 * Rot = Tn
      MultiHomoMtrx(tmpT, Tn, Rot);  //      ROt = inv(T0) * Tn
      // get rotation matrix parameters
      double k_vec[3], omega;
      GetRotKparam(Rot, k_vec, &omega);
      std::cout << "k_vector=[" << k_vec[0] << ", " << k_vec[1] << ", " << k_vec[2] << "], omega=" << omega << std::endl;
      double domega = 0.05; // [rad]
      if(fabs(omega/domega) > secNum){
	secNum = fabs(omega/domega);
      }
      domega = omega / secNum;

      double dr[3];
      for(i=0; i<3; i++){
	dr[i] =(LastPos[i] - CurPos[i]) / secNum;
	RefPos[i] = CurPos[i];
      }

      //      std::cout << "section number = " << secNum << std::endl; // for debug

      for(i=0; i<JOINTNUM; i++){
	RefJPos[i] = CurJPos[i];
      }
      
      CopyHomoMtrx(R[JOINTNUM-1], Tw0);
      CopyHomoMtrx(R[JOINTNUM-1], Tw);
      //      std::cout << "Tw=" << Tw[0][3] << ", " << Tw[1][3] << ", " << Tw[2][3] << std::endl;

      for(i=0; i<secNum; i++){
	ForwardKinematics(RefJPos, Ti, Ri);
	CalcJacobi(Ri, Jw);
	
	// PosLog << Ri[JOINTNUM][0][3] << " " << Ri[JOINTNUM][1][3] << " " << Ri[JOINTNUM][2][3] << " ";
	// std::cout << "[" << i << "] x=[" << Ri[JOINTNUM][0][3] << ", " << Ri[JOINTNUM][1][3] << ", " << Ri[JOINTNUM][2][3] << "] xr=[";

	CopyHomoMtrx(Tw, Tw0);
	CalcRotK(k_vec, domega*(i+1), Rot);
	MultiHomoMtrx(T0, Rot, tmpT);
	
	for(j=0; j<3; j++){
	  RefPos[j] = RefPos[j] + dr[j];
	  tmpT[j][3] = RefPos[j];
	}
	MultiHomoMtrx(tmpT, inv_Te, Tw);
	
	CalcInvHomoMtrx(Tw0, tmpT);
	MultiHomoMtrx(Tw, tmpT, dTw);  // dTw * Tw0 = Tw, dTw = Tw * inv(Tw0)
	dww[0] = dTw[2][1];
	dww[1] = dTw[0][2];
	dww[2] = dTw[1][0];
	for(j=0; j<3; j++){
	  drw[j] = Tw[j][3] - Tw0[j][3];
	  for(k=0; k<JOINTNUM-3; k++){
	    J11[j][k] = Jw[j][k];
	  }
	}
	// // for debug
	// std::cout << "["<< i << "] drw=[" << drw[0] << ", " << drw[1] << ", " << drw[2] << "]";
	// std::cout << ", dww=[" << dww[0] << ", " << dww[1] << ", " << dww[2] << "]" << std::endl;	
	SolveSysEq(J11, drw, dtheta);
	
	double tmp = 0.0;
	for(j=0; j<3; j++){
	  tmp = 0.0;
	  for(k=0; k<JOINTNUM-3; k++){
	    tmp += Jw[j+3][k] * dtheta[k];
	  }
	  dww[j] = dww[j] - tmp;
	}
	// // for debug
	// std::cout << "dww=[" << dww[0] << ", " << dww[1] << ", " << dww[2] << "]" << std::endl;
	for(j=0; j<3; j++){
	  for(k=0; k<3; k++){
	    J22[j][k] = Jw[j+3][k+JOINTNUM-3];
	  }
	}
	SolveSysEq33(J22, dww, &dtheta[3]);
	
	// // for debug
	// std::cout << "["<< i << "] dtheta=[";	
	// for(j=0; j<JOINTNUM; j++){
	//   std::cout << dtheta[j] << ", ";	
	// }
	// std::cout << "]"<< std::endl;

	
	//	std::cout << "RefJpos=[";
	for(j=0; j<JOINTNUM; j++){
	  RefJPos[j] = RefJPos[j] + dtheta[j];
	  //	  std::cout << RefJPos[j] << ", ";
	}
	//	std::cout << "]" << std::endl;

	for(j=0; j<JOINTNUM; j++){
	  JPosLog << RefJPos[j] << " ";
	}
	JPosLog << std::endl;

	PosLog << RefPos[0] << " " << RefPos[1] << " " << RefPos[2] << std::endl;	

      }
      for(i=0; i<JOINTNUM; i++){
	LastJPos[i] = RefJPos[i];
      } 
      //      std::cout << "Finish ! " << std::endl;
      ForwardKinematics(LastJPos, Ti, Ri);
      // // for debug
      // CalcPositionVector(Ri[JOINTNUM], zero);
      // std::cout << "[" << i << "] x=[" << zero[0] << ", " << zero[1] << ", " << zero[2] << ", "
      // 		<< zero[3] << ", " << zero[4] << ", " << zero[5] << "]" << std::endl;
    }
  }


  
  if(iFlgCalcTrajParam){
    CalcTrajParam();
    iFlgCalcTrajParam = 0;
  }  

  t = t + dT;

  if(t > TIME){
    for(i=0; i<JOINTNUM; i++){
      RefJPos[i] = LastJPos[i];
      RefJVel[i] = 0.0;
      //      RefJAcc[i] = 0.0;
    }
  }else{
    // generate target velocity
    double t2 = t * t;
    double t3 = t2 * t;
    double t4 = t3 * t;
    double t5 = t4 * t;

    for(i=0; i<JOINTNUM; i++){
      RefJPos[i] = a[i][4] * t5 + a[i][3] * t4 + a[i][2] * t3 + a[i][1] * t2 + a[i][0] * t + InitJPos[i];
      RefJVel[i] = 5*a[i][4] * t4 + 4*a[i][3] * t3 + 3*a[i][2] * t2 + 2*a[i][1] * t + a[i][0];
    }
  }

  for(i=0; i<JOINTNUM; i++){
    m_TargetVelocity.data[i] = (RefJPos[i] - CurJPos[i])*10.0;
  }
  m_TargetVelocityOut.write();
  
  JPosLog << NowTime << " " << dT << " " << t << " ";
  JVelLog << NowTime << " " << dT << " " << t << " ";

  for(i=0; i<JOINTNUM; i++){
    JPosLog << RefJPos[i] << " ";
    JVelLog << RefJVel[i] << " ";

  }
  for(i=0; i<JOINTNUM; i++){
    JPosLog << CurJPos[i] << " ";
    JVelLog << CurJVel[i] << " ";

  }
  JPosLog << std::endl;
  JVelLog << std::endl;

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t MotionGenerator::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t MotionGenerator::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t MotionGenerator::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t MotionGenerator::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t MotionGenerator::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

 // calculation of forward kinematics for RH
void MotionGenerator::ForwardKinematics(double angle[JOINTNUM], double T0[JOINTNUM][3][4], double R0[JOINTNUM+1][3][4])
{

  int i,j;
  
  T0[0][0][0] = cos(angle[0]);  T0[0][0][1] = -sin(angle[0]); T0[0][0][2] = 0.0;  T0[0][0][3] = 0.0;
  T0[0][1][0] = sin(angle[0]);  T0[0][1][1] = cos(angle[0]);  T0[0][1][2] = 0.0;  T0[0][1][3] = 0.0;
  T0[0][2][0] = 0.0;            T0[0][2][1] = 0.0;            T0[0][2][2] = 1.0;  T0[0][2][3] = L[0];

  T0[1][0][0] = sin(angle[1]);  T0[1][0][1] = cos(angle[1]);  T0[1][0][2] = 0.0;  T0[1][0][3] = 0.0;
  T0[1][1][0] = 0.0;            T0[1][1][1] = 0.0;            T0[1][1][2] = 1.0;  T0[1][1][3] = 0.0;
  T0[1][2][0] = cos(angle[1]);  T0[1][2][1] = -sin(angle[1]); T0[1][2][2] = 0.0;  T0[1][2][3] = L[1];  

  T0[2][0][0] = cos(angle[2]);  T0[2][0][1] = -sin(angle[2]); T0[2][0][2] = 0.0;  T0[2][0][3] = L[2];
  T0[2][1][0] = sin(angle[2]);  T0[2][1][1] = cos(angle[2]);  T0[2][1][2] = 0.0;  T0[2][1][3] = 0.0;
  T0[2][2][0] = 0.0;            T0[2][2][1] = 0.0;            T0[2][2][2] = 1.0;  T0[2][2][3] = 0.0;

  T0[3][0][0] = 0.0;            T0[3][0][1] = 0.0;            T0[3][0][2] = 1.0;  T0[3][0][3] = L[3];
  T0[3][1][0] = cos(angle[3]);  T0[3][1][1] = -sin(angle[3]); T0[3][1][2] = 0.0;  T0[3][1][3] = 0.0;
  T0[3][2][0] = sin(angle[3]);  T0[3][2][1] = cos(angle[3]);  T0[3][2][2] = 0.0;  T0[3][2][3] = 0.0;

  T0[4][0][0] = sin(angle[4]);  T0[4][0][1] = cos(angle[4]);  T0[4][0][2] = 0.0;  T0[4][0][3] = 0.0;
  T0[4][1][0] = 0.0;            T0[4][1][1] = 0.0;            T0[4][1][2] = 1.0;  T0[4][1][3] = 0.0;
  T0[4][2][0] = cos(angle[4]);  T0[4][2][1] = -sin(angle[4]); T0[4][2][2] = 0.0;  T0[4][2][3] = L[4];

  T0[5][0][0] = 0.0;            T0[5][0][1] = 0.0;            T0[5][0][2] = 1.0;  T0[5][0][3] = L[5];
  T0[5][1][0] = cos(angle[5]);  T0[5][1][1] = -sin(angle[5]); T0[5][1][2] = 0.0;  T0[5][1][3] = 0.0;
  T0[5][2][0] = sin(angle[5]);  T0[5][2][1] = cos(angle[5]);  T0[5][2][2] = 0.0;  T0[5][2][3] = 0.0;

  for(i=0; i<3; i++){
    for(j=0; j<4; j++){
      R0[0][i][j] = T0[0][i][j];
    }
  }

  for(i=0; i<JOINTNUM-1; i++){
    MultiHomoMtrx(R0[i], T0[i+1], R0[i+1]);
  }
  
  MultiHomoMtrx(R0[JOINTNUM-1], Te, R0[JOINTNUM]);
}

void MotionGenerator::CalcHomoMtrx(double pose[6], double trns[3][4]){
  int i;
  double sa, ca, sb, cb, sc, cc;
  sa = sin(pose[3]);   ca = cos(pose[3]); // Z
  sb = sin(pose[4]);   cb = cos(pose[4]); // Y
  sc = sin(pose[5]);   cc = cos(pose[5]); // X

  for(i=0; i<3; i++){
    trns[i][3] = pose[i];
  }
  trns[0][0] = ca*cb;
  trns[1][0] = sa*cb;
  trns[2][0] = -sb;

  trns[0][1] = ca*sb*sc - sa*cc;
  trns[1][1] = sa*sb*sc + ca*cc;
  trns[2][1] = cb*sc;

  trns[0][2] = ca*sb*cc + sa*sc;
  trns[1][2] = sa*sb*cc - ca*sc;
  trns[2][2] = cb*cc;  

}

// calculation jacobi matrix of wrist position
void MotionGenerator::CalcJacobi(double R[JOINTNUM+1][3][4], double J[6][JOINTNUM]){
  int i,j;
  double Z[JOINTNUM][3], P[JOINTNUM][3], Jw11T[JOINTNUM][3];
  
  for(i=0; i<JOINTNUM; i++){
    for(j=0; j<3; j++){
      Z[i][j] = R[i][j][2];
      P[i][j] = R[JOINTNUM-1][j][3] - R[i][j][3];
      J[j+3][i] = Z[i][j];
    }
  }

  for(i=0; i<JOINTNUM; i++){
    CalcOuterProduct(Z[i], P[i], Jw11T[i]);
    for(j=0; j<3; j++){
      J[j][i] = Jw11T[i][j];
    }
  }

  for(i=0; i<3; i++){
    for(j=0; j<3; j++){
      J[j][i+(JOINTNUM-3)] = 0.0;
    }
  }
  // // for debug
  // std::cout << "Jw=[";
  // for(i=0; i<6; i++){
  //   for(j=0; j<JOINTNUM; j++){
  //     std::cout << J[i][j] << ", ";
  //   }
  //   std::cout << "]" << std::endl;
  // } 
}

void MotionGenerator::CalcPositionVector(double Rn[3][4], double pose[6]){
  int i;
  // position
  for(i=0; i<3; i++){
    pose[i] = Rn[i][3];
  }
  // posture    // ZYX Eular
  if(fabs(Rn[1][0])<CUTOFF_VALUE){
    Rn[1][0] = 0.0;
  }
  if(fabs(Rn[0][0])<CUTOFF_VALUE){
    Rn[0][0] = 0.0;
  }
  double gamma = atan2(Rn[1][0], Rn[0][0]);
  pose[5] = gamma;
  
  double c_gamma = cos(gamma);
  double s_gamma = sin(gamma);

  if(fabs(Rn[2][0])<CUTOFF_VALUE){
    Rn[2][0] = 0.0;
  }
  pose[4] = atan2(-Rn[2][0], (c_gamma*Rn[0][0]+s_gamma*Rn[1][0]));

  if(fabs(Rn[0][2])<CUTOFF_VALUE){
    Rn[0][2] = 0.0;
  }
  if(fabs(Rn[1][2])<CUTOFF_VALUE){
    Rn[1][2] = 0.0;
  }
  if(fabs(Rn[0][1])<CUTOFF_VALUE){
    Rn[0][1] = 0.0;
  }
  if(fabs(Rn[1][1])<CUTOFF_VALUE){
    Rn[1][1] = 0.0;
  }
  pose[3] = atan2((s_gamma*Rn[0][2]-c_gamma*Rn[1][2]),(-s_gamma*Rn[0][1]+c_gamma*Rn[1][1]));
}

void MotionGenerator::CalcPostureVector(double R[3][3], double posture[3]){

  // posture    // X-Z-X rotation
  if(fabs(R[1][2])<CUTOFF_VALUE){
    R[1][2] = 0.0;
  }
  if(fabs(R[2][2])<CUTOFF_VALUE){
    R[2][2] = 0.0;
  }
  posture[0] = atan2(R[1][2], R[2][2]);
  
  double c0 = cos(posture[0]);
  double s0 = sin(posture[0]);

  if(fabs(R[0][2])<CUTOFF_VALUE){
    R[0][2] = 0.0;
  }
  posture[1] = atan2(R[0][2], (c0*R[1][2] + s0*R[2][2]));

  if(fabs(R[1][0])<CUTOFF_VALUE){
    R[1][0] = 0.0;
  }
  if(fabs(R[2][0])<CUTOFF_VALUE){
    R[2][0] = 0.0;
  }
  if(fabs(R[1][1])<CUTOFF_VALUE){
    R[1][1] = 0.0;
  }
  if(fabs(R[2][1])<CUTOFF_VALUE){
    R[2][1] = 0.0;
  }
  posture[2] = atan2((-s0*R[1][1] + c0*R[2][1]), (-s0*R[1][0] + c0*R[2][0]));
}

int MotionGenerator::GetRotKparam(double rotK[3][4], double kr[3], double *theta){
  double phi, tmp, sqrt_tmp;
  double r32_r23, r13_r31, r21_r12;

  for(int i=0; i<3; i++){
    if(fabs(rotK[i][3])<CUTOFF_VALUE){
      rotK[i][3] = 0.0;
    }
  }
  
  r32_r23 = rotK[2][1]-rotK[1][2];
  r13_r31 = rotK[0][2]-rotK[2][0];
  r21_r12 = rotK[1][0]-rotK[0][1];

  tmp = r32_r23 * r32_r23 + r13_r31 * r13_r31 + r21_r12 * r21_r12;
  sqrt_tmp = sqrt(tmp);

  phi = atan2(sqrt_tmp, (rotK[0][0]+rotK[1][1]+rotK[2][2]-1.0));
  *theta = phi;

  double sp = sin(phi);
  double cp = cos(phi);

  if(sp < CUTOFF_VALUE){
    if(phi < CUTOFF_VALUE){      // phi = 0.0
      kr[0] = 0.0;
      kr[1] = 0.0;
      kr[2] = 1.0;
      return(-1);
    }else
    if(phi < 0.1){
      // normalization
      kr[0] = r32_r23/sqrt_tmp;
      kr[1] = r13_r31/sqrt_tmp;
      kr[2] = r21_r12/sqrt_tmp;
    }else{
      // phi = 180.0
      if(rotK[0][0]>rotK[1][1] && rotK[0][0]>rotK[2][2]){
	kr[0] = GetSgn(r32_r23)*sqrt((rotK[0][0]-cp)/(1.0-cp));
	if(abs(kr[0])>CUTOFF_VALUE){
	  kr[1] = (rotK[1][0]+rotK[0][1])/(2.0*kr[0]*(1.0-cp));
	  kr[2] = (rotK[2][0]+rotK[0][2])/(2.0*kr[0]*(1.0-cp));
	}else{
	  std::cout << "Error! Can't get RotK parameters::kr[0]=0.0" << std::endl;
	  return(-100);
	}
      }else
      if(rotK[1][1]>rotK[2][2]){
	kr[1] = GetSgn(r13_r31)*sqrt((rotK[1][1]-cp)/(1.0-cp));
	if(abs(kr[1])>CUTOFF_VALUE){
	  kr[0] = (rotK[1][0]+rotK[0][1])/(2.0*kr[1]*(1.0-cp));
	  kr[2] = (rotK[2][1]+rotK[1][2])/(2.0*kr[1]*(1.0-cp));
	}else{
	  std::cout << "Error! Can't get RotK parameters::kr[1]=0.0" << std::endl;
	  return(-100);
	}
      }else{
	kr[2] = GetSgn(r21_r12)*sqrt((rotK[2][2]-cp)/(1.0-cp));
	if(abs(kr[2])>CUTOFF_VALUE){
	  kr[0] = (rotK[2][0]+rotK[0][2])/(2.0*kr[2]*(1.0-cp));
	  kr[1] = (rotK[2][1]+rotK[1][2])/(2.0*kr[2]*(1.0-cp));
	}else{
	  std::cout << "Error! Can't get RotK parameters::kr[2]=0.0" << std::endl;
	  return(-100);
	}
      }
      return(-2);
    }
  // }else
  // if(cos(phi) < -0.785){
  //   // phi = 180.0
  //   if(rotK[0][0]>rotK[1][1] && rotK[0][0]>rotK[2][2]){
  //     kr[0] = GetSgn(r32_r23)*sqrt((rotK[0][0]-cp)/(1.0-cp));
  //     if(abs(kr[0])>CUTOFF_VALUE){
  // 	kr[1] = (rotK[1][0]+rotK[0][1])/(2.0*kr[0]*(1.0-cp));
  // 	kr[2] = (rotK[2][0]+rotK[0][2])/(2.0*kr[0]*(1.0-cp));
  //     }else{
  // 	std::cout << "Error! Can't get RotK parameters::kr[0]=0.0" << std::endl;
  // 	return(-100);
  //     }
  //   }else
  //   if(rotK[1][1]>rotK[2][2]){
  //     kr[1] = GetSgn(r13_r31)*sqrt((rotK[1][1]-cp)/(1.0-cp));
  //     if(abs(kr[1])>CUTOFF_VALUE){
  // 	kr[0] = (rotK[1][0]+rotK[0][1])/(2.0*kr[1]*(1.0-cp));
  // 	kr[2] = (rotK[2][1]+rotK[1][2])/(2.0*kr[1]*(1.0-cp));
  //     }else{
  // 	std::cout << "Error! Can't get RotK parameters::kr[1]=0.0" << std::endl;
  // 	return(-100);
  //     }
  //   }else{
  //     kr[2] = GetSgn(r21_r12)*sqrt((rotK[2][2]-cp)/(1.0-cp));
  //     if(abs(kr[2])>CUTOFF_VALUE){
  // 	kr[0] = (rotK[2][0]+rotK[0][2])/(2.0*kr[2]*(1.0-cp));
  // 	kr[1] = (rotK[2][1]+rotK[1][2])/(2.0*kr[2]*(1.0-cp));
  //     }else{
  // 	std::cout << "Error! Can't get RotK parameters::kr[2]=0.0" << std::endl;
  // 	return(-100);
  //     }
  //   }
  //   return(-2);
  }else{
    kr[0] = r32_r23/(2.0*sp);
    kr[1] = r13_r31/(2.0*sp);
    kr[2] = r21_r12/(2.0*sp);
    
    return(0);
  }
  return(0);
}

// calculation of homomatrix from rotation matrix parameters
void MotionGenerator::CalcRotK(double kr[3], double theta, double rotK[3][4]){

  double s = sin(theta);
  double c = cos(theta);
  double v = 1-c;
  
  rotK[0][0] = kr[0]*kr[0]*v + c;
  rotK[1][0] = kr[0]*kr[1]*v + kr[2]*s;
  rotK[2][0] = kr[0]*kr[2]*v - kr[1]*s;

  rotK[0][1] = kr[1]*kr[0]*v - kr[2]*s;
  rotK[1][1] = kr[1]*kr[1]*v + c;
  rotK[2][1] = kr[1]*kr[2]*v + kr[0]*s;

  rotK[0][2] = kr[2]*kr[0]*v + kr[1]*s;
  rotK[1][2] = kr[2]*kr[1]*v - kr[0]*s;
  rotK[2][2] = kr[2]*kr[2]*v + c;

}

void MotionGenerator::MultiHomoMtrx(double A[3][4], double B[3][4], double C[3][4]){

  for(int i = 0; i < 3; i++){
    for(int j = 0; j < 3; j++){

      C[i][j] = 0.0;

      for(int k = 0; k < 3; k++){
	C[i][j] += A[i][k] * B[k][j];
      }
      if(fabs(C[i][j])<CUTOFF_VALUE){
	C[i][j] = 0.0;
      }
    }
    C[i][3] = 0.0;
    for(int k = 0; k < 3; k++){
      C[i][3] += A[i][k] * B[k][3];
    }
    C[i][3] += A[i][3];

    if(fabs(C[i][3])<CUTOFF_VALUE){
      C[i][3] = 0.0;
    }
  }
}
double MotionGenerator::CalcDistance(double Pa[6], double Pb[6]){
  double tmp = (Pa[0] - Pb[0])*(Pa[0] - Pb[0]) + (Pa[1] - Pb[1])*(Pa[1] - Pb[1]) + (Pa[2] - Pb[2])*(Pa[2] - Pb[2]);
  tmp = sqrt(tmp);

  return tmp;

}

void MotionGenerator::CalcOuterProduct(double a[3], double b[3], double c[3]){
  c[0] = a[1]*b[2] - a[2]*b[1];
  c[1] = a[2]*b[0] - a[0]*b[2];
  c[2] = a[0]*b[1] - a[1]*b[0];
}

void MotionGenerator::CalcInvHomoMtrx(double mat_a[3][4], double mat_b[3][4]){
  int i,j;
  for(i=0; i<3; i++){
    for(j=0; j<3; j++){
      mat_b[i][j] = mat_a[j][i];
      if(fabs(mat_b[i][j])<CUTOFF_VALUE){
	mat_b[i][j]=0.0;
      }
    }
  }
  double inv_pose[3];
  for(i=0; i<3; i++){
    inv_pose[i] = -mat_a[i][3];
  }
  // p'=(R)T*(-p)
  for(i=0; i<3; i++){
    mat_b[i][3] = 0.0;
    for(j=0; j<3; j++){
      mat_b[i][3] += mat_a[j][i]*inv_pose[j];
    }
    if(fabs(mat_b[i][3])<CUTOFF_VALUE){
      mat_b[i][3]=0.0;
    }
  }
}

void MotionGenerator::CopyHomoMtrx(double mat_a[3][4], double mat_b[3][4]){
  int i,j;
  for(i=0; i<3; i++){
    for(j=0; j<4; j++){
      mat_b[i][j] = mat_a[i][j];
    }
  }
}
int MotionGenerator::GetSgn(double x){
  if(x < 0.0){
    return(-1);
  }else{
    return(1);
  }
}

void MotionGenerator::SolveSysEq(double A[3][JOINTNUM-3], double b[3], double x[JOINTNUM-3]){
  int i,j,k;
  int N = JOINTNUM-3;
  double A_T[N][3], A_TA[N][N], c[N];
  double L[N][N], U[N][N];  // for LU decomposition
  double y[N];
  for(i=0; i<N; i++){
    for(j=0; j<3; j++){
      A_T[i][j] = A[j][i];
    }
  }
  for(i=0; i<N; i++){

    for(j=0; j<N; j++){
      L[i][j] = 0.0;
      U[i][j] = 0.0;
      if(i==j){
	L[i][j] = 1.0;
	U[i][j] = 1.0;
      }
      A_TA[i][j] = 0.0;
      for(k=0; k<3; k++){
	A_TA[i][j] += A_T[i][k] * A[k][j];
      }
      if(fabs(A_TA[i][j])<CUTOFF_VALUE){
	A_TA[i][j] = 0.0;
      }
    }
    c[i] = 0.0;
    for(k=0; k<3; k++){
      c[i] += A_T[i][k] * b[k];
    }
  }
  // std::cout << "c=AT*b=[";
  // for(i=0; i<N; i++){
  //   std::cout << c[i] << ", ";
  // }
  // std::cout << "]" << std::endl;
  
  // LU decomposition
  for(j=0; j<N;j++){
    for(i=0; i<j+1; i++){
      U[i][j] = A_TA[i][j];

      for(k=0; k<i; k++){
	U[i][j] = U[i][j] - L[i][k] * U[k][j];
      }
      if(i==j && fabs(U[i][i])<CUTOFF_VALUE){
	U[i][i] = 0.0001; // magic number
      }
    }
    for(i=j+1; i<N; i++){
      L[i][j] = A_TA[i][j];
      for(k=0; k<j; k++){
	L[i][j] = L[i][j] - L[i][k] * U[k][j];
      }
      L[i][j] = L[i][j] / U[j][j];
    }
  }

  for(i=0; i<N;i++){
    y[i] = c[i];
    for(j=0; j<i; j++){
      y[i] = y[i] - y[j] * L[i][j];
    }
  }

  // std::cout << "y=LT*c=[";
  // for(i=0; i<N; i++){
  //   std::cout << y[i] << ", ";
  // }
  // std::cout << "]" << std::endl;
  
  for(i=N-1; i>=0; i--){
    x[i] = y[i];
    for(j=i+1; j<N; j++){
      x[i] = x[i] - U[i][j] * x[j];
    }
    x[i] = x[i] / U[i][i];
  }

  // std::cout << "c=UT*y=[";
  // for(i=0; i<N; i++){
  //   std::cout << x[i] << ", ";
  // }
  // std::cout << "]" << std::endl;
  
  // // for debug
  // std::cout << "A, A_TA, L, U" << std::endl;
  // for(i=0; i<N; i++){
  //   std::cout << "[";
  //   for(j=0; j<N; j++){
  //     std::cout << A[i][j] << ", ";
  //   }
  //   std::cout << "] [";    
  //   for(j=0; j<N; j++){
  //     std::cout << A_TA[i][j] << ", ";
  //   }
  //   std::cout << "] [";
  //   for(j=0; j<N; j++){
  //     std::cout << L[i][j] << ", ";
  //   }
  //   std::cout << "] [";
  //   for(j=0; j<N; j++){
  //     std::cout << U[i][j] << ", ";
  //   }
  //   std::cout << "]" << std::endl;
  // }

}

void MotionGenerator::SolveSysEq33(double A[3][3], double b[3], double x[3]){
  int i,j,k;
  double A_T[3][3], A_TA[3][3], c[3];
  double L[3][3], U[3][3];  // for LU decomposition
  double y[3];
  for(i=0; i<3; i++){
    for(j=0; j<3; j++){
      A_T[i][j] = A[j][i];
    }
  }
  for(i=0; i<3; i++){

    for(j=0; j<3; j++){
      L[i][j] = 0.0;
      U[i][j] = 0.0;
      if(i==j){
	L[i][j] = 1.0;
	U[i][j] = 1.0;
      }
      A_TA[i][j] = 0.0;
      for(k=0; k<3; k++){
	A_TA[i][j] += A_T[i][k] * A[k][j];
      }
      if(fabs(A_TA[i][j])<CUTOFF_VALUE){
	A_TA[i][j] = 0.0;
      }
    }
    c[i] = 0.0;
    for(k=0; k<3; k++){
      c[i] += A_T[i][k] * b[k];
    }
  }
  // LU decomposition
  for(j=0; j<3;j++){
    for(i=0; i<j+1; i++){
      U[i][j] = A_TA[i][j];

      for(k=0; k<i; k++){
	U[i][j] = U[i][j] - L[i][k] * U[k][j];
      }
      if(i==j && fabs(U[i][i])<CUTOFF_VALUE){
	U[i][i] = 0.0001; // magic number
      }
    }
    for(i=j+1; i<3; i++){
      L[i][j] = A_TA[i][j];
      for(k=0; k<j; k++){
	L[i][j] = L[i][j] - L[i][k] * U[k][j];
      }
      L[i][j] = L[i][j] / U[j][j];
    }
  }

  for(i=0; i<3;i++){
    y[i] = c[i];
    for(j=0; j<i; j++){
      y[i] = y[i] - y[j] * L[i][j];
    }
  }

  for(i=2; i>=0; i--){
    x[i] = y[i];
    for(j=i+1; j<3; j++){
      x[i] = x[i] - U[i][j] * x[j];
    }
    x[i] = x[i] / U[i][i];
  }
}
// calculation of target trajectory parameters
void MotionGenerator::CalcTrajParam(){
  
  double b[JOINTNUM][3];

  for(int i=0; i<JOINTNUM; i++){
    InitJPos[i] = CurJPos[i];
    //      InitJVel[i] = CurJVel[i];
    InitJVel[i] = 0.0;
    //      InitJAcc[i] = CurJAcc[i];
    InitJAcc[i] = 0.0;

    a[i][1] = 0.5*InitJAcc[i];
    a[i][0] = InitJVel[i];
    b[i][0] = -InitJAcc[i]/TIME;
    b[i][1] = b[i][0]-InitJVel[i]/(TIME*TIME);
    b[i][2] = (LastJPos[i]-InitJPos[i]-InitJVel[i]*TIME-a[i][1]*TIME*TIME)/(TIME*TIME*TIME);

    a[i][4] = (0.5*b[i][0] - 3*b[i][1] + 6*b[i][2])/(TIME*TIME);
    a[i][3] = (-1*b[i][0] - 7*b[i][1] - 15*b[i][2])/TIME;
    a[i][2] = 0.5*b[i][0] - 4*b[i][1] + 10*b[i][2];
      
    t = 0.0;
    std::cout << "NO." << i+1 << ": InitJPos=" << InitJPos[i];
    //    std::cout << ": InitVel=" << InitJVel[i] << ": InitAcc=" << InitJAcc[i] << std::endl;
    std::cout << "a5=" << a[i][4] << ", a4=" << a[i][3] << ", a3=" << a[i][2] << ", a2=" << a[i][1] << ", a1=" << a[i][0] << std::endl;
    std::cout << "b1=" << b[i][0] << ", b2=" << b[i][1] << ", b3=" << b[i][2] << std::endl;
  }
}

extern "C"
{
 
  void MotionGeneratorInit(RTC::Manager* manager)
  {
    coil::Properties profile(motiongenerator_spec);
    manager->registerFactory(profile,
                             RTC::Create<MotionGenerator>,
                             RTC::Delete<MotionGenerator>);
  }
  
};



