// -*- C++ -*-
/*!
 * @file  SetTargetPosition.cpp * @brief target data setting component * $Date$ 
 *
 * $Id$ 
 */
#include "SetTargetPosition.h"

// Module specification
// <rtc-template block="module_spec">
static const char* settargetposition_spec[] =
  {
    "implementation_id", "SetTargetPosition",
    "type_name",         "SetTargetPosition",
    "description",       "target data setting component",
    "version",           "1.0",
    "vendor",            "Junichiro Ooga, TOSHIBA",
    "category",          "RobotPlanner",
    "activity_type",     "EVENT_DRIVEN",
    "kind",              "DataFowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    ""
  };
// </rtc-template>

SetTargetPosition::SetTargetPosition(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_TargetPositionOut("TargetPosition", m_TargetPosition)

    // </rtc-template>
{
}

SetTargetPosition::~SetTargetPosition()
{
}


RTC::ReturnCode_t SetTargetPosition::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers

  // Set OutPort buffer
  addOutPort("TargetPosition", m_TargetPositionOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable

  // </rtc-template>
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t SetTargetPosition::onFinalize()
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t SetTargetPosition::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t SetTargetPosition::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t SetTargetPosition::onActivated(RTC::UniqueId ec_id)
{
  m_TargetPosition.data.length(6);
  for(int i=0; i<6; i++){
    m_TargetPosition.data[i] = 0.0;
  }
  m_TargetPositionOut.write();
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SetTargetPosition::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t SetTargetPosition::onExecute(RTC::UniqueId ec_id)
{
  double tmp[6];
  std::cout << "Input Target Position [m/deg]" << std::endl;
  std::cout << "x [m] = ";
  std::cin >> tmp[0];
  std::cout << "y [m] = ";
  std::cin >> tmp[1];
  std::cout << "z [m] = ";
  std::cin >> tmp[2];  
  std::cout << "a [deg] = ";
  std::cin >> tmp[3];
  std::cout << "b [deg] = ";
  std::cin >> tmp[4];
  std::cout << "c [deg] = ";
  std::cin >> tmp[5];

  for(int i=0; i<3; i++){
    m_TargetPosition.data[i] = tmp[i];
    m_TargetPosition.data[i+3] = tmp[i+3] * M_PI / 180.0;
  }
  
  m_TargetPositionOut.write();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SetTargetPosition::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t SetTargetPosition::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t SetTargetPosition::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t SetTargetPosition::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t SetTargetPosition::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


extern "C"
{
 
  void SetTargetPositionInit(RTC::Manager* manager)
  {
    coil::Properties profile(settargetposition_spec);
    manager->registerFactory(profile,
                             RTC::Create<SetTargetPosition>,
                             RTC::Delete<SetTargetPosition>);
  }
  
};



